#include <EEPROM.h>
#include <Adafruit_GFX.h>
#include <Adafruit_PCD8544.h>

#define ENC_A 2 //А энкодера
#define ENC_B 3 //В энкодера
#define ENC_BUTTON 4 //кнопка энкодера
#define FOCUS_OUT 5 //Пин фокуса
#define SHUTTER_OUT 6 //Пин шаттера
#define BACKLIGHT 12 //Управление подсветкой
#define BUZZER_OUT 13 //Пин подключен к пищалке
#define AUX A5 //Пин расширения
#define V_BAT A7 //пин контроля батареи

Adafruit_PCD8544 display = Adafruit_PCD8544(7, 8, 9, 10, 11); //SCLK, DIN, D/C, CS, RST

unsigned long currentMillis = 0; //текущее время
unsigned long previousMillis = 0; //предыдущее время
unsigned long estMillis = 0; //расчетное время

bool batErr = false; //ошибка батареи
int batLevel = 0; //напряжение батареи

bool tamper = false; //контроль подсветки
unsigned int timeCount = 0; //переменная подсветки
int dsp = 0; //контраст дисплея

long interval = 0; //интервал снимков
int count = 0; //количество снимков
int lmt = 0; //лимит снимков
float fps = 0; //кадров в секунду
float clipDuration = 0; //расчетное время клипа

unsigned int cycle; //переменная промежутков в цикле
volatile long newPosition = 0; //переменная валкодера
unsigned int choice = 0; //переменная выбора
bool set = false; //Режим установки параметра
bool settMode = false; //Режим настройки
bool start = false; //пуск / стоп

const unsigned char PROGMEM batt[] =
{ B11111111, B11111111, B11111111, B11110000,
  B10000000, B00000000, B00000000, B00011100,
  B10000000, B00000000, B00000000, B00010100,
  B10000000, B00000000, B00000000, B00010100,
  B10000000, B00000000, B00000000, B00010100,
  B10000000, B00000000, B00000000, B00010100,
  B10000000, B00000000, B00000000, B00011100,
  B11111111, B11111111, B11111111, B11110000
};

//чтение EEPROM long
unsigned long EEPROM_ulong_read(int addr)
{
  byte raw[4];
  for (byte i = 0; i < 4; i++) raw[i] = EEPROM.read(addr + i);
  unsigned long &num = (unsigned long&)raw;
  return num;
}

//запись EEPROM long
void EEPROM_ulong_write(int addr, unsigned long num)
{
  byte raw[4];
  (unsigned long&)raw = num;
  for (byte i = 0; i < 4; i++) EEPROM.update(addr + i, raw[i]);
}

void EEPROMread()
{
  interval = EEPROM_ulong_read(0);
  count = EEPROM_ulong_read(4);
  fps = EEPROM_ulong_read(8);
  dsp = EEPROM_ulong_read(12);
}

void EEPROMwrite()
{
  EEPROM_ulong_write(0 , interval);
  if (count <= 0)EEPROM_ulong_write(4, count);
  EEPROM_ulong_write(8 , fps);
  EEPROM_ulong_write(12 , dsp);
}

void backLight_on() //включение подсветки
{
  timeCount = 0;
  tamper = true;
}

void backlightTimer() //автовыключение подсветки
{
  if (tamper)
  {
    digitalWrite(BACKLIGHT, HIGH);
    timeCount++;
  }
  if (timeCount >= 300) //подобрать по времени подсветки
  {
    digitalWrite(BACKLIGHT, LOW);
    tamper = false;
    timeCount = 0;
  }
}

void batLoop() //цикл мониторинга батареи
{
  float Vmin = 6; //нижний порог напряжения
  float Vmax = 8.4; //верхний порог напряжения
  float A_K = 0.2; //Коэффициент сглаживания
  float k = 100 / (Vmax - Vmin); //коэффициент расчета процентов
  float voltage = analogRead(V_BAT) * (5.0 / 1023) * 2; //расчет напряжения
  static float A_y; //Выходное значение
  A_y = A_y - A_K * (A_y - voltage); //Сглаживание значения заряда батареи
  batLevel = (A_y - Vmin) * k; //перевод в проценты
  if (A_y < Vmin + 0.1 || A_y > Vmax + 0.1) batErr = true;
  else batErr = false;
  batLevel = max(batLevel, 0);
  batLevel = min(batLevel, 100);
}

void LCDloop() //цикл дисплея
{
  unsigned long time = currentMillis / 1000;
  unsigned long _interval = interval;
  unsigned int _choice = choice;
  int _batLevel = batLevel;
  bool _set = set;
  display.clearDisplay();

  if (!settMode)
  {
    unsigned long est = 0;
    unsigned int showBat = 1;

    if (start) est = _interval - (currentMillis - estMillis) / 1000;
    else
    {
      est = _interval;
      lmt = count;
    }
    if (count >= 0) clipDuration = count / fps;
    else clipDuration = lmt * -1 / fps;

    if (batErr) if (time & 1) showBat = 0;

    //отрисивка заряда батареи
    display.drawBitmap(0, 0,  batt, 30, 8, showBat); //рисуем батарейку
    if (_batLevel > 13) display.fillRect(2, 2, 4, 4, 1); //6.3v
    if (_batLevel > 50) display.fillRect(7, 2, 4, 4, 1); //7.2v
    if (_batLevel > 58) display.fillRect(12, 2, 4, 4, 1); //7.4v
    if (_batLevel > 65) display.fillRect(17, 2, 4, 4, 1); //7.55v
    if (_batLevel > 80) display.fillRect(22, 2, 4, 4, 1); //7.9v

    //отрисивка времени работы
    display.setCursor(35, 0);
    display.setTextColor(WHITE, BLACK);
    if (time / 60 / 60 < 10) display.print ("0");
    display.print (time / 60 / 60);
    display.print (":");
    if (time / 60 % 60 < 10) display.print ("0");
    display.print ((time / 60) % 60);
    display.print (":");
    if (time % 60 < 10) display.print ("0");
    display.print (time % 60);
    display.setTextColor(BLACK , WHITE);

    //отрисивка расчетного времени до следующего кадра
    display.setCursor(0, 10);
    display.print ("Next: ");
    if (est / 60 / 60 < 10) display.print ("0");
    display.print (est / 60 / 60);
    display.print (":");
    if (est / 60 % 60 < 10) display.print ("0");
    display.print ((est / 60) % 60);
    display.print (":");
    if (est % 60 < 10) display.print ("0");
    display.print (est % 60);

    //отрисивка счетчика кадров
    display.setCursor(0, 19);
    display.print ("Shot: ");
    display.print (count);

    //отрисивка расчетного времени клипа
    display.setCursor(0, 28);
    display.print ("Clip: ");
    display.print (clipDuration, 1);
    display.print ("s");

    //отрисивка кнопок "start/set"
    display.drawRect(0, 38, 38, 10, BLACK);
    display.drawRect(45, 38, 38, 10, BLACK);

    display.setCursor(1, 39);
    if (_choice == 1) display.setTextColor(WHITE, BLACK);
    else display.setTextColor(BLACK, WHITE);
    display.print (" Sett ");

    display.setCursor(46, 39);
    if (_choice == 0) display.setTextColor(WHITE, BLACK);
    else display.setTextColor(BLACK, WHITE);
    if (start) display.print (" Stop ");
    else display.print ("  Go  ");
  }
  else
  {
    unsigned int SH = 1;
    if (time & 1) SH = 0;

    //отрисивка интервала
    display.setCursor(0, 0);
    display.setTextColor(BLACK , WHITE);
    display.print ("Int: ");

    if (_choice == 2)
    {
      if (_set) display.setTextColor(SH, !SH);
      else display.setTextColor(WHITE, BLACK);
    }
    if (_interval / 60 / 60 < 10) display.print ("0");
    display.print (_interval / 60 / 60);
    display.setTextColor(BLACK, WHITE);
    display.print (":");

    if (_choice == 3)
    {
      if (_set) display.setTextColor(SH, !SH);
      else display.setTextColor(WHITE, BLACK);
    }
    if (_interval / 60 % 60 < 10) display.print ("0");
    display.print ((_interval / 60) % 60);
    display.setTextColor(BLACK, WHITE);
    display.print (":");

    if (_choice == 4)
    {
      if (_set) display.setTextColor(SH, !SH);
      else display.setTextColor(WHITE, BLACK);
    }
    if (_interval % 60 < 10) display.print ("0");
    display.print (_interval % 60);
    display.setTextColor(BLACK, WHITE);

    //отрисивка счетчика кадров
    display.setCursor(0, 9);
    display.print ("Lmt: ");
    if (_choice == 5)
    {
      if (_set) display.setTextColor(SH, !SH);
      else display.setTextColor(WHITE, BLACK);
    }
    if (count > 0) display.print ("0");
    else display.print (count * -1);
    display.setTextColor(BLACK, WHITE);
    display.print (" sht");

    //отрисивка кадров в секунду
    display.setCursor(0, 18);
    display.print ("Fps: ");
    if (_choice == 6)
    {
      if (_set) display.setTextColor(SH, !SH);
      else display.setTextColor(WHITE, BLACK);
    }
    display.print (fps, 0);
    display.setTextColor(BLACK, WHITE);

    //отрисивка контраста
    display.setCursor(0, 27);
    display.print ("Dsp: ");
    if (_choice == 7)
    {
      if (_set) display.setTextColor(SH, !SH);
      else display.setTextColor(WHITE, BLACK);
    }
    display.print (dsp);
    display.setTextColor(BLACK, WHITE);


    //отрисивка кнопки "Back"
    display.drawRect(0, 38, 38, 10, BLACK);
    display.setCursor(1, 39);
    if (_choice == 1) display.setTextColor(WHITE, BLACK);
    else display.setTextColor(BLACK, WHITE);
    display.print (" Back ");
  }
  display.display();
}

void doEncoderA() // функция действия вызываемое прерыванием
{
  if (digitalRead(ENC_A) == digitalRead(ENC_B)) newPosition++;
  else newPosition--;
}

void encoderLoop() //цикл обработки валкодера
{
  static volatile long oldPosition;
  unsigned long curPosition = newPosition;

  if (curPosition != oldPosition)
  {
    backLight_on();
    tone(BUZZER_OUT, 4500, 10);
    if (!settMode)
    {
      if (choice == 0) choice = 1;
      else choice = 0;
    }
    else
    {
      choice = constrain(choice, 1, 7);
      if (curPosition > oldPosition)
      {
        if (set)
        {
          if (choice == 2)interval += 3600; //прибавляем часы
          if (choice == 3)interval += 60; //прибавляем минуты
          if (choice == 4)interval++; //прибавляем секунды
          if (choice == 5)
          {
            if (count <= 0)count--; //прибавляем лимит кадров
            else count = 0;
          }
          if (choice == 6)fps++;
          if (choice == 7)
          {
            dsp++;
            display.setContrast(dsp);
          }
        }
        else
        {
          if (choice < 7)choice++;
          else choice = 1;
        }
      }
      if (curPosition < oldPosition)
      {
        if (set)
        {
          if (choice == 2)interval -= 3600; //уменьшаем часы
          if (choice == 3)interval -= 60; //уменьшаем минуты
          if (choice == 4)interval--; //уменьшаем секунды
          if (choice == 5)
          {
            if (count < 0)count++; //уменьшаем лимит кадров
            else count = 0;
          }
          if (choice == 6)fps--;
          if (choice == 7)
          {
            dsp--;
            display.setContrast(dsp);
          }
        }
        else
        {
          if (choice > 1)choice--;
          else choice = 7;
        }
      }
    }
    oldPosition = curPosition;
  }
  interval = constrain(interval, 1, 86400);
  dsp = constrain(dsp, 1, 100);
}

unsigned int duratation(byte button) //получаем время до отпускания кнопки
{
  unsigned int result = 0;
  while (digitalRead(button) == LOW)
  {
    delay(80); //тут шаг можно отрегулировать на рельном девайсе
    if (digitalRead(button) == LOW) result += 100;
    if (result > 1100) return 1100; //таймаут нажатия
  }
  return result;
}

void buttonLoop() //цикл кнопки
{
  if (digitalRead(ENC_BUTTON) == LOW)
  {
    backLight_on();
    unsigned int duration = duratation(ENC_BUTTON);
    if (duration > 1000) //длинное нажатие
    {
      if (choice == 0)
      {
        tone(BUZZER_OUT, 4500, 200);
        if (!start)
        {
          EEPROMwrite();
          if (count > 0)count = 0;
          estMillis = currentMillis;
          previousMillis = currentMillis;
          cycle = 0;
        }
        start = !start;
        LCDloop();
        if (!start)
        {
          digitalWrite(FOCUS_OUT, LOW);
          digitalWrite(SHUTTER_OUT, LOW);
          digitalWrite(AUX, LOW);
        }
        while (digitalRead(ENC_BUTTON) == LOW) delay(1); //ждём пока отпустят кнопку
      }
    }
    else if (duration > 10)  //короткое нажатие
    {
      tone(BUZZER_OUT, 4500, 20);
      buttonClick();
      while (digitalRead(ENC_BUTTON) == LOW) delay(1); //ждём пока отпустят кнопку
    }
  }
}

void buttonClick() //обработка режима кнопки и перехода по настройкам
{
  if (choice == 0)
  {
    if (!start)manualShot();
    return;
  }
  else if (choice == 1)
  {
    settMode = !settMode;
    return;
  }
  else
  {
    set = !set;
    return;
  }
}

void manualShot()
{
  digitalWrite(SHUTTER_OUT, HIGH);
  digitalWrite(AUX, HIGH);
  delay(200);
  digitalWrite(SHUTTER_OUT, LOW);
  digitalWrite(AUX, LOW);
  count++;
}

void mainLoop() //основной цикл работы
{
  static int limitShot;
  int _batLevel = batLevel;

  if (currentMillis - previousMillis >= interval * 500)
  {
    previousMillis = currentMillis;
    cycle++;
    if (cycle == 1)
    {
      //действие между кадрами
      if (_batLevel < 5)tone(BUZZER_OUT, 4500, 250);
      digitalWrite(FOCUS_OUT, HIGH);
    }
    if (cycle >= 2)
    {
      estMillis = currentMillis;
      digitalWrite(SHUTTER_OUT, HIGH);
      digitalWrite(AUX, HIGH);
      limitShot = count;
      count++;
      cycle = 0;
    }
  }
  if (currentMillis - previousMillis >= 200)
  {
    if (digitalRead(SHUTTER_OUT) == HIGH) digitalWrite(FOCUS_OUT, LOW);
    digitalWrite(SHUTTER_OUT, LOW);
    digitalWrite(AUX, LOW);
    if (limitShot < 0)
    {
      if (count == 0)
      {
        limitShot = 0;
        start = false;
        //"Done" beep
        tone(BUZZER_OUT, 4500, 250);
        delay(500);
        tone(BUZZER_OUT, 4500, 250);
        delay(500);
        tone(BUZZER_OUT, 4500, 250);
        //
      }
    }
  }
}

void defaultSett()
{
  interval = 1;
  count = 0;
  fps = 25;
  dsp = 50;
  EEPROMwrite();
  tone(BUZZER_OUT, 4500, 250);
  delay(500);
  tone(BUZZER_OUT, 4500, 250);
  delay(500);
  tone(BUZZER_OUT, 4500, 250);
}

void setup()
{
  pinMode(AUX, OUTPUT);
  pinMode(FOCUS_OUT, OUTPUT);
  pinMode(SHUTTER_OUT, OUTPUT);
  pinMode(BUZZER_OUT, OUTPUT);
  pinMode(BACKLIGHT, OUTPUT);
  pinMode(ENC_A, INPUT_PULLUP);
  pinMode(ENC_B, INPUT_PULLUP);
  pinMode(ENC_BUTTON, INPUT_PULLUP);
  if (digitalRead(ENC_BUTTON) == LOW) defaultSett();
  EEPROMread();
  attachInterrupt(0, doEncoderA, CHANGE);
  display.begin();
  display.clearDisplay();
  display.setContrast(dsp);
}

void loop()
{
  currentMillis = millis(); //основное время
  if (start) mainLoop();
  else encoderLoop();
  batLoop();
  LCDloop();
  buttonLoop();
  backlightTimer();
}
